@@ -17,9 +17,6 @@ def handle_request(host_queues, requests, gpu_request):
         destination = gpu_request['rank']
         host = gpu_request['host']
     else:
-        if gpu_request == "pass":
-            ranks_exit_request.put("exit")
-
         parts = gpu_request.split(':')
         if len(parts) != 3:
             logging.warn("Invalid message %s", gpu_request)
@@ -29,6 +26,7 @@ def handle_request(host_queues, requests, gpu_request):
         destination = parts[2]
 
     logging.debug("[%s][%s] Got gpu request: %s", host, rank, gpu_request)
+    logging.debug("[%s][%s] Got gpu request[host_queues]: %s", host, rank,host_queues)
     gpu_queue = host_queues[host]
 
     if destination:
@@ -66,30 +64,61 @@ class gpu:
         def load_gpus():
             from blazer.hpc.mpi import parallel, pipeline, partial as p, scatter, where, select, filter, rank, size
 
-            if rank != 0:
-                return []
-
             def get_gpus():
                 from blazer.hpc.mpi import rank
                 import platform
-                logging.info("GETTING GPU for rank[%s]",rank)
-                gpus = main()
-                for gpu in gpus:
-                    gpu['host'] = platform.node()
-                    gpu['rank'] = rank
-                return gpus
+                logging.debug("GETTING GPU for rank[%s]",rank)
+                try:
+                    gpus = main()
+                    for gpu in gpus:
+                        gpu['host'] = platform.node()
+                        gpu['rank'] = rank
+                    return gpus
+                except:
+                    return []
 
             gpu_calls = [p(get_gpus) for i in range(1,size)]
             gpus = parallel(gpu_calls)
-            print("GET_GPUS:",gpus)
-            return gpus
+            if not gpus:
+                gpus = []
+            gpulist = [gpu for gpu in gpus if len(gpu) > 0]
+
+            logging.debug("GOT GPUs for rank[%s] %s",rank, gpulist)
+            return gpulist
             
 
         self.GPUS = []
 
         try:
+            from pydash import flatten
             if rank == 0:
                 self.GPUS = load_gpus()
+                self.GPUS = flatten(self.GPUS)
+                self.gpuranks = len(self.GPUS)
+                for gpu in self.GPUS:
+                    if gpu['host'] not in self.host_queues:
+                        self.host_queues[gpu['host']] = SimpleQueue()
+                    self.host_queues[gpu['host']].put(gpu)
+
+                #comm.Barrier()
+            
+            else:
+                import platform
+                #print("WAITING FOR MASTER TO GATHER GPU DATA")
+                #comm.Barrier()
+                logging.info("[%s][%s] GPU context init. Checking my GPUS", host,rank)
+                try:
+                    gpus = main()
+                    for gpu in gpus:
+                        gpu['host'] = platform.node()
+                        gpu['rank'] = rank
+                    self.GPUS = flatten(gpus)
+                except Exception as ex:
+                    logging.error(ex)
+                    self.GPUS = []
+
+                logging.info("GOT GPUs for [%s][%s] %s",host,rank, self.GPUS)
+            
         except:
             import traceback
             print(traceback.format_exc())
@@ -100,10 +129,6 @@ class gpu:
     def __enter__(self, *args, **kwargs) -> Any: 
         logging.debug("[%s][%s] GPU Context enter",host,rank)
 
-        if len(self.GPUS) == 0:
-            logging.info("[%s][%s] I don't have any GPUS so just exiting",host,rank)
-            return None
-
         while True:
 
             if rank == 0:
@@ -112,15 +137,17 @@ class gpu:
                 logging.debug("size is %s and ranks_exit_request.qsize()+1 is %s", size, ranks_exit_request.qsize()+1)
 
                 logging.debug("total_released %s size %s",self.total_released, size-1)
-                
+                logging.debug("gpuranks is %s and ranks_exit_request.qsize()+1 is %s", self.gpuranks, ranks_exit_request.qsize()+1)
                 # If the # of rank exist requests + 1 (master) equals the total number of ranks
                 # Then master node can exit. All ranks have reported in
-                if size == ranks_exit_request.qsize()+1:
+                if self.gpuranks == ranks_exit_request.qsize()+1:
                     logging.debug("MASTER FINISHED: total_released = %s",self.total_released)
                     break
                     
                 # Wait for GPU requests on tag 1. Block until we get a message
+                logging.debug("MASTER recv")
                 gpu_request = comm.recv(tag=1)
+                logging.debug("MASTER got gpu_request")
 
                 logging.debug("[%s][%s] Master got request from rank %s", host, rank, gpu_request)
                         
@@ -157,16 +184,20 @@ class gpu:
                     # Handle the message
                     handle_request(self.host_queues, self.requests, gpu_request)
             else:
+                if len(self.GPUS) == 0:
+                    logging.debug("[%s][%s] I don't have any GPUS so just exiting",host,rank)
+                    break
+
                 logging.debug("[%s][%s] Sending gpu request",host,rank)
 
                 # Request a GPU from master node. 
                 comm.send(f"gpu:{host}:{rank}", dest=0, tag=1)
                 
                 # Block until we get one: NOTE: What if server finishes and this receive is never fulfilled?
-                logging.debug("[%s][%s] Waiting for gpu",host,rank)
+                logging.debug("[%s][%s] RECV Waiting for gpu",host,rank)
                 self.using_gpu = gpu = comm.recv(source=0, tag=1)
 
-                logging.debug("[%s][%s] Received GPU[%s]",host,rank, gpu)
+                logging.debug("[%s][%s] RECV Received GPU[%s]",host,rank, gpu)
                 cuda.select_device(gpu['id'])
                 self.has_cuda = True
                 # Resume context in app code
@@ -190,7 +221,7 @@ class gpu:
                 logging.debug("[%s][%s] GPU Context exit: sending gpu back to master GPU %s",host,rank, self.using_gpu)
                 comm.send(self.using_gpu, dest=0, tag=1)
                 logging.debug("[%s][%s] GPU Context exit: released GPU %s",host,rank, self.using_gpu)
-                ranks_exit_request.put("exit")
+                #ranks_exit_request.put("exit")
             else:
                 # Master node exiting GPU context. Receive any stuck messages?
                 logging.debug("[%s][%s] MASTER GPU Context exit",host,rank)
